package obo;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;

/* this class provides the method get_tree_lca corresponding to the LCA.m MATLAB function 
 * OUTPUT: 
	its generates the following files
 	TREE_LCA, matrix matrice containing for each couple of terms with at least one annotation the number of genes annotated with the minimum subsumer
	terms_for_LCA, list of terms with at least one annotation
 	mapping_ID_indici, hashtable containing the association term ID and the corresponding index in the TREE_LCA matrix
	mapping_Indici_ID, hashtable opposite to mapping_ID_indici
 	termini_ordinati, GO term in  terms_for_LCA increasingly sort by ID
 	TREE_P_max, maximum value in GO_GENE, i.d. the maximum number of annotations for a GO term weighted by updegree (number of parents)
*/
public class Get_TREE_LCA {
	
	public void get_tree_lca(String Path_version,String Path){
		System.out.println("Calcolo TREE_LCA...");
		/* prelevo i termini in GO_GENE_NoBranch(TREE_P_NoBranch) che hanno una valore>0(quindi che hanno almeno un gene annotato),
		* questi termini serviranno per costruire la matrice TREE_LCA(vedi LCA.m nel codice di Paccanaro). 
		* La matrice TREE_LCA contiene per ogni
		* coppia di termini i-j il numero minimo di geni annotati con l'antenato in comune con minor numero
		* di geni annotati.NOTA: nel calcolo degli antenati di un termine, vengono esclusi i termini che 
		* non hanno geni annotati,ossia che non appartengono alla lista dei temini che hanno 
		* GO_GENE_NoBranch(TREE_P_NoBranch)>0.Inoltre, tra gli antenati di un termine figura anche il termine stesso.
		*/
		
		//carico go_ontology
		OBOReader go_ontology = null;
		ObjectInputStream ois7;
		System.out.println("start load go_ontology...");
		try{
			ois7 = new ObjectInputStream(new FileInputStream(Path_version+"/go_ontology"));
		    go_ontology = (OBOReader) ois7.readObject();
		}
		catch(Exception e){
			System.out.println("error during loading go_ontology");
			e.printStackTrace();
		}
		System.out.println("done!");
	    //----------------------
		
		//carico GO_GENE (TREE_P)
		Hashtable<String, Double> GO_GENE = new Hashtable<String, Double>();
	    ObjectInputStream ois9;
		System.out.println("start load GO_GENE...");
		try{
			ois9 = new ObjectInputStream(new FileInputStream(Path+"/GO_GENE"));
			GO_GENE = (Hashtable<String, Double>) ois9.readObject();
		}
		catch(Exception e){
			System.out.println("error during loading GO_GENE");
		    e.printStackTrace();
		}
		System.out.println("done!");
	    //-------------------------
		
		//carico GO_GENE_NoBranch (TREE_P_NoBranch)
	    Hashtable<String, Double> GO_GENE_NoBranch = new Hashtable<String, Double>();
	    ObjectInputStream ois0;
		System.out.println("start load GO_GENE_NoBranch...");
		try{
			ois0 = new ObjectInputStream(new FileInputStream(Path+"/GO_GENE_NoBranch"));
		    GO_GENE_NoBranch = (Hashtable<String, Double>) ois0.readObject();
		}
		catch(Exception e){
			System.out.println("error during loading GO_GENE_NoBranch");
		    e.printStackTrace();
		}
		System.out.println("done!");
		//--------------------------
		
		ArrayList<String> terms_for_LCA = new ArrayList<String>();//termini che hanno valore in GO_GENE_NoBranch>0.
		Hashtable<String, Double> TREE_P_filtrato = new Hashtable<String, Double>();//conterr il sottoinsieme dei termini in GO_GENE_NoBranch che hanno valore>0. 
		for(String st : GO_GENE_NoBranch.keySet()){//scorre TREE_P_NoBranch
			if(GO_GENE_NoBranch.get(st)>0.0) {
				TREE_P_filtrato.put(st, GO_GENE_NoBranch.get(st));
				terms_for_LCA.add(st);
			}
		}
		System.out.println("dimensione terms_for_LCA = "+terms_for_LCA.size());
		
		//ordino gli ID in base al loro valore numerico
		ArrayList<String> termini_ordinati = new ArrayList<String>();
		Hashtable<String, Integer> num = new Hashtable<String, Integer>();
		Hashtable<Integer,String> num_inv = new Hashtable<Integer, String>();
		for(String te:terms_for_LCA){
			int num_te = Integer.parseInt(te.substring(3));
			num.put(te, num_te);
			num_inv.put(num_te, te);
		}
		int[] for_sort = new int[terms_for_LCA.size()];
		int ci=0;
		for(String term:num.keySet()){
			for_sort[ci] = num.get(term);
			ci++;
		}
		Arrays.sort(for_sort);
		for(int p=0;p<for_sort.length;p++){
			termini_ordinati.add(num_inv.get(for_sort[p]));
		}
		
		/* in terms_processed manterremo i termini man mano processati durante il calcolo di TREE_LCA.
		 * In questo modo una volta calcolato TREE_LCA di i-j avremo anche calcolato TREE_LCA di j-i. */
		ArrayList<String> terms_processed = new ArrayList<String>();		
		System.out.println("Calcolo matrice TREE_LCA...(l'operazione richiede qualche minuto)");
		double[][] TREE_LCA2 = new double[terms_for_LCA.size()][terms_for_LCA.size()];
		/* Le seguenti hashtable contengono per ogni termini in terms_for_LCA(ossia termini che hanno TREE_P_NoBranch>0)
		 * il corrispondente indice intero e viceversa. */
		Hashtable<String, Integer> mapping_ID_Indici = new Hashtable<String, Integer>();//ID_TERMINE-indice
	    Hashtable<Integer, String> mapping_Indici_ID = new Hashtable<Integer, String>();//indice-ID_TERMINE
	    int index=0;
	    for(String di: termini_ordinati){
	        mapping_ID_Indici.put(di, index);
	        mapping_Indici_ID.put(index, di);
	        index++;
	    }
	    
	    //Calcolo TREE_LCA
		for(String ter: termini_ordinati){//scorro termini con TREE_P_NoBranch>0
			//list_id.add(ter);PER TEST SUL NUMERO ANTENATI
			int ter_ind = mapping_ID_Indici.get(ter);//prendo indice termine
			String antenati_1 = new String();//prendo antenati primo termine
			antenati_1 = go_ontology.getAncestorList(go_ontology.getTerm(ter), antenati_1);
			ArrayList<String> lista_ante_1 = new ArrayList<String>();
			String[] a1;
			a1 = antenati_1.split("\n");
			for(String d:a1){
				if(!terms_for_LCA.contains(d))continue;//filtro solo gli antenati che hanno TREE_P_NoBranch>0
		        lista_ante_1.add(d);
			}
			//nume_ante.add(lista_ante_1.size()); PER TEST SUL NUMERO ANTENATI
			
			/* scorro tutti gli altri termini e di ognuno prendo gli antenati e li confronto
			 * con quelli del termine precedente, trovo gli antenati in comune e per ciascuno
			 * di questi prelevo il numero di geni annotato indicato in GO_GENE_NoBranch, quindi
			 * selezione il numero minimo di geni annotati e inserisco il valore in TREE_LCA2.
			 * NOTA: una cella della matrice TREE_LCA2(t1,t2) contiene il num min di geni annotato per
			 * l'antenato in comune con num minimo di geni annotati
			 */
			for(String ter2: termini_ordinati){
				int ter2_ind = mapping_ID_Indici.get(ter2);//prendo indice termine di confronto
				/* controllo che nella matrice TREE_LCA2 non ci sia gi l'info relativa al termine
				 * di confronto. questo vorrebbe dire che TREE_LCA2 relativo ai 2 termini  
				 * gia stato calcolato
				 */
				if(terms_processed.contains(ter2))continue;
				String antenati_2 = new String();//prendo antenati
				antenati_2 = go_ontology.getAncestorList(go_ontology.getTerm(ter2), antenati_2);
				ArrayList<String> lista_ante_2 = new ArrayList<String>();
				String[] a2;
				a2 = antenati_2.split("\n");
				for(String h:a2){
					if(!terms_for_LCA.contains(h))continue;//filtro solo gli antenati che hanno TREE_P_NoBranch>0
					lista_ante_2.add(h);
				}	
				//confronto gli antenati
				ArrayList<String> ante_t1_t2 = new ArrayList<String>();
			    for(String t : lista_ante_1){//scorro gli antenati di t1
			        for(String l : lista_ante_2){//scorro gli antenati di t2
			        	if(t.compareTo(l)==0){//trovato elemento in comune
			        		ante_t1_t2.add(t);
			        		break;
			        	}
			        }
			    }
			    
			    Double num_min_geni = Integer.MAX_VALUE+0.0;//num_min_geni conterr il numero minimo di geni annotati con uno degli antenati comuni
			    for(String ante : ante_t1_t2){/* prendo l'antenato in comune con minor num di geni annotati */
			    	if(TREE_P_filtrato.get(ante)<num_min_geni) num_min_geni=TREE_P_filtrato.get(ante);
			    }
			    
				/*Inserisco TREE_LCA2(t1,t2) nella matrice TREE_LCA2.
				 * Vengono inserite 2 tupla, una con t1,t2 e l'altra con t2,t1.
				 */
			    TREE_LCA2[ter_ind][ter2_ind]=num_min_geni;
			    TREE_LCA2[ter2_ind][ter_ind]=num_min_geni;
			}//chiude for che scorre tutti gli altri termini
			terms_processed.add(ter);//aggiungo il termine alla lista dei termini processati
		}//chiude for che scorre tutti i termini
		System.out.println("fatto!");
		
		/* per i valori in TREE_LCA<0.1 inserisco 0.1 (operazione che viene effettuata nel file LCA.m di Paccanaro) */
		for(int r=0;r<TREE_LCA2.length;r++){//scorro righe
			for(int c=0;c<TREE_LCA2[0].length;c++){//scorro colonne
				if(TREE_LCA2[r][c]<0.1)TREE_LCA2[r][c]=0.1;
			}
		}
		
		/* Calcolo TREE_P_max come avviene in GOSIM_Pre.m del codice di Paccanaro. Mi servir per calcolare Resnik. */
	    Double TREE_P_max = 0.0;//il TREE_P qui  GO_GENE, quindi TREE_P pesato
	    for(String yui : GO_GENE.keySet()){
	        if(GO_GENE.get(yui)>TREE_P_max) TREE_P_max = GO_GENE.get(yui); 
	    }
		
		StoreFile(Path+"/TREE_LCA", TREE_LCA2);
		StoreFile(Path+"/terms_for_LCA",terms_for_LCA);
		StoreFile(Path+"/mapping_ID_Indici", mapping_ID_Indici);
		StoreFile(Path+"/mapping_Indici_ID", mapping_Indici_ID);
		StoreFile(Path+"/termini_ordinati", termini_ordinati);
		StoreFile(Path+"/TREE_P_max", TREE_P_max);
		
		System.out.println("Fine calcolo TREE_LCA!");
	}
	
	public static void StoreFile(String FileName, Object structure){
		System.out.println("Storing "+FileName+"...");
		ObjectOutputStream oss;
        try{
            oss = new ObjectOutputStream(new FileOutputStream(FileName));
            oss.writeObject(structure);
            oss.close();
        }
        catch(Exception e1){
                System.out.println("error during store "+FileName);
                e1.printStackTrace();
        }
        System.out.println("Done!");
	}
	
	public void get_tree_lca_version(String Date,String Path){
		System.out.println("Calcolo TREE_LCA_"+Date+"...");
		/* prelevo i termini in GO_GENE_NoBranch(TREE_P_NoBranch) che hanno una valore>0(quindi che hanno almeno un gene annotato),
		* questi termini serviranno per costruire la matrice TREE_LCA(vedi LCA.m nel codice di Paccanaro). 
		* La matrice TREE_LCA contiene per ogni
		* coppia di termini i-j il numero minimo di geni annotati con l'antenato in comune con minor numero
		* di geni annotati.NOTA: nel calcolo degli antenati di un termine, vengono esclusi i termini che 
		* non hanno geni annotati,ossia che non appartengono alla lista dei temini che hanno 
		* GO_GENE_NoBranch(TREE_P_NoBranch)>0.Inoltre, tra gli antenati di un termine figura anche il termine stesso.
		*/
		
		//carico go_ontology
		OBOReader go_ontology = null;
		ObjectInputStream ois7;
		System.out.println("start load go_ontology_"+Date+"...");
		try{
			ois7 = new ObjectInputStream(new FileInputStream(Path+"sim_calc/go_ontology_"+Date));
		    go_ontology = (OBOReader) ois7.readObject();
		}
		catch(Exception e){
			System.out.println("error during loading go_ontology_"+Date);
			e.printStackTrace();
		}
		System.out.println("done!");
	    //----------------------
		
		//carico GO_GENE (TREE_P)
		Hashtable<String, Double> GO_GENE = new Hashtable<String, Double>();
	    ObjectInputStream ois9;
		System.out.println("start load GO_GENE_"+Date+"...");
		try{
			ois9 = new ObjectInputStream(new FileInputStream(Path+"sim_calc/GO_GENE_"+Date));
			GO_GENE = (Hashtable<String, Double>) ois9.readObject();
		}
		catch(Exception e){
			System.out.println("error during loading GO_GENE_"+Date);
		    e.printStackTrace();
		}
		System.out.println("done!");
	    //-------------------------
		
		//carico GO_GENE_NoBranch (TREE_P_NoBranch)
	    Hashtable<String, Double> GO_GENE_NoBranch = new Hashtable<String, Double>();
	    ObjectInputStream ois0;
		System.out.println("start load GO_GENE_NoBranch_"+Date+"...");
		try{
			ois0 = new ObjectInputStream(new FileInputStream(Path+"sim_calc/GO_GENE_NoBranch_"+Date));
		    GO_GENE_NoBranch = (Hashtable<String, Double>) ois0.readObject();
		}
		catch(Exception e){
			System.out.println("error during loading GO_GENE_NoBranch_"+Date);
		    e.printStackTrace();
		}
		System.out.println("done!");
		//--------------------------
		
		ArrayList<String> terms_for_LCA = new ArrayList<String>();//termini che hanno valore in GO_GENE_NoBranch>0.
		Hashtable<String, Double> TREE_P_filtrato = new Hashtable<String, Double>();//conterr il sottoinsieme dei termini in GO_GENE_NoBranch che hanno valore>0. 
		for(String st : GO_GENE_NoBranch.keySet()){//scorre TREE_P_NoBranch
			if(GO_GENE_NoBranch.get(st)>0.0) {
				TREE_P_filtrato.put(st, GO_GENE_NoBranch.get(st));
				terms_for_LCA.add(st);
			}
		}
		System.out.println("dimensione terms_for_LCA_"+Date+" = "+terms_for_LCA.size());
		
		//ordino gli ID in base al loro valore numerico
		ArrayList<String> termini_ordinati = new ArrayList<String>();
		Hashtable<String, Integer> num = new Hashtable<String, Integer>();
		Hashtable<Integer,String> num_inv = new Hashtable<Integer, String>();
		for(String te:terms_for_LCA){
			int num_te = Integer.parseInt(te.substring(3));
			num.put(te, num_te);
			num_inv.put(num_te, te);
		}
		int[] for_sort = new int[terms_for_LCA.size()];
		int ci=0;
		for(String term:num.keySet()){
			for_sort[ci] = num.get(term);
			ci++;
		}
		Arrays.sort(for_sort);
		for(int p=0;p<for_sort.length;p++){
			termini_ordinati.add(num_inv.get(for_sort[p]));
		}
		
		/* in terms_processed manterremo i termini man mano processati durante il calcolo di TREE_LCA.
		 * In questo modo una volta calcolato TREE_LCA di i-j avremo anche calcolato TREE_LCA di j-i. */
		ArrayList<String> terms_processed = new ArrayList<String>();		
		System.out.println("Calcolo matrice TREE_LCA...(l'operazione richiede qualche minuto)");
		double[][] TREE_LCA2 = new double[terms_for_LCA.size()][terms_for_LCA.size()];
		/* Le seguenti hashtable contengono per ogni termini in terms_for_LCA(ossia termini che hanno TREE_P_NoBranch>0)
		 * il corrispondente indice intero e viceversa. */
		Hashtable<String, Integer> mapping_ID_Indici = new Hashtable<String, Integer>();//ID_TERMINE-indice
	    Hashtable<Integer, String> mapping_Indici_ID = new Hashtable<Integer, String>();//indice-ID_TERMINE
	    int index=0;
	    for(String di: termini_ordinati){
	        mapping_ID_Indici.put(di, index);
	        mapping_Indici_ID.put(index, di);
	        index++;
	    }
	    
	    //Calcolo TREE_LCA
		for(String ter: termini_ordinati){//scorro termini con TREE_P_NoBranch>0
			//list_id.add(ter);PER TEST SUL NUMERO ANTENATI
			int ter_ind = mapping_ID_Indici.get(ter);//prendo indice termine
			String antenati_1 = new String();//prendo antenati primo termine
			antenati_1 = go_ontology.getAncestorList(go_ontology.getTerm(ter), antenati_1);
			ArrayList<String> lista_ante_1 = new ArrayList<String>();
			String[] a1;
			a1 = antenati_1.split("\n");
			for(String d:a1){
				if(!terms_for_LCA.contains(d))continue;//filtro solo gli antenati che hanno TREE_P_NoBranch>0
		        lista_ante_1.add(d);
			}
			//nume_ante.add(lista_ante_1.size()); PER TEST SUL NUMERO ANTENATI
			
			/* scorro tutti gli altri termini e di ognuno prendo gli antenati e li confronto
			 * con quelli del termine precedente, trovo gli antenati in comune e per ciascuno
			 * di questi prelevo il numero di geni annotato indicato in GO_GENE_NoBranch, quindi
			 * selezione il numero minimo di geni annotati e inserisco il valore in TREE_LCA2.
			 * NOTA: una cella della matrice TREE_LCA2(t1,t2) contiene il num min di geni annotato per
			 * l'antenato in comune con num minimo di geni annotati
			 */
			for(String ter2: termini_ordinati){
				int ter2_ind = mapping_ID_Indici.get(ter2);//prendo indice termine di confronto
				/* controllo che nella matrice TREE_LCA2 non ci sia gi l'info relativa al termine
				 * di confronto. questo vorrebbe dire che TREE_LCA2 relativo ai 2 termini  
				 * gia stato calcolato
				 */
				if(terms_processed.contains(ter2))continue;
				String antenati_2 = new String();//prendo antenati
				antenati_2 = go_ontology.getAncestorList(go_ontology.getTerm(ter2), antenati_2);
				ArrayList<String> lista_ante_2 = new ArrayList<String>();
				String[] a2;
				a2 = antenati_2.split("\n");
				for(String h:a2){
					if(!terms_for_LCA.contains(h))continue;//filtro solo gli antenati che hanno TREE_P_NoBranch>0
					lista_ante_2.add(h);
				}	
				//confronto gli antenati
				ArrayList<String> ante_t1_t2 = new ArrayList<String>();
			    for(String t : lista_ante_1){//scorro gli antenati di t1
			        for(String l : lista_ante_2){//scorro gli antenati di t2
			        	if(t.compareTo(l)==0){//trovato elemento in comune
			        		ante_t1_t2.add(t);
			        		break;
			        	}
			        }
			    }
			    
			    Double num_min_geni = Integer.MAX_VALUE+0.0;//num_min_geni conterr il numero minimo di geni annotati con uno degli antenati comuni
			    for(String ante : ante_t1_t2){/* prendo l'antenato in comune con minor num di geni annotati */
			    	if(TREE_P_filtrato.get(ante)<num_min_geni) num_min_geni=TREE_P_filtrato.get(ante);
			    }
			    
				/*Inserisco TREE_LCA2(t1,t2) nella matrice TREE_LCA2.
				 * Vengono inserite 2 tupla, una con t1,t2 e l'altra con t2,t1.
				 */
			    TREE_LCA2[ter_ind][ter2_ind]=num_min_geni;
			    TREE_LCA2[ter2_ind][ter_ind]=num_min_geni;
			}//chiude for che scorre tutti gli altri termini
			terms_processed.add(ter);//aggiungo il termine alla lista dei termini processati
		}//chiude for che scorre tutti i termini
		System.out.println("fatto!");
		
		/* per i valori in TREE_LCA<0.1 inserisco 0.1 (operazione che viene effettuata nel file LCA.m di Paccanaro) */
		for(int r=0;r<TREE_LCA2.length;r++){//scorro righe
			for(int c=0;c<TREE_LCA2[0].length;c++){//scorro colonne
				if(TREE_LCA2[r][c]<0.1)TREE_LCA2[r][c]=0.1;
			}
		}
		
		/* Calcolo TREE_P_max come avviene in GOSIM_Pre.m del codice di Paccanaro. Mi servir per calcolare Resnik. */
	    Double TREE_P_max = 0.0;//il TREE_P qui  GO_GENE, quindi TREE_P pesato
	    for(String yui : GO_GENE.keySet()){
	        if(GO_GENE.get(yui)>TREE_P_max) TREE_P_max = GO_GENE.get(yui); 
	    }
		
		StoreFile("TREE_LCA_"+Date, TREE_LCA2);
		StoreFile("terms_for_LCA_"+Date,terms_for_LCA);
		StoreFile("mapping_ID_Indici_"+Date, mapping_ID_Indici);
		StoreFile("mapping_Indici_ID_"+Date, mapping_Indici_ID);
		StoreFile("termini_ordinati_"+Date, termini_ordinati);
		StoreFile("TREE_P_max_"+Date, TREE_P_max);
		
		System.out.println("Fine calcolo TREE_LCA_"+Date+"!");
	}

}
