package obo;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Hashtable;

import sim_calc.FileIn;

/* The method storeOBO parse the OBO file. Corresponds to the MATLAB function get_Trees.m
 * INPUT:
	path_to_OBO_file, path to the OBO file
 * OUTPUT
	it generates the following files:
  			go_ontology, parsed ontology
  			TreeBP, Up_BP, Down_BP  , for each term containing the info about the annotated genes, the term parents and the term children
  			TreeCC Up_CC, Down_CC,
  			TreeMF, Up_MF, Down_MF,
  			ALLGOLIST, contains the info relative to all the GO terms.
 */
public class ParsingOBO {
	
	public void storeOBO (String versione,String path_to_OBO_file){
		System.out.println("path OBO File = "+path_to_OBO_file);
		System.out.println("Parsing obo file...");
		OBOReader go_ontology = new OBOReader(path_to_OBO_file);
		go_ontology.buildHier();
		System.out.println("End!");
		
		/* creo le strutture dati relative ad ogni dominio (BP,CC,MF).
		 * TreeXX conterr� le info relative ai termini del dominio XX.
		 * UpXX conterr� le info relative ai padri di ogni termini in XX.
		 * DownXX conterr� le info relative ai figli di ogni termini in XX.
		 * ALLGOLIST conterr� le info relative a tutti i termini. */
		ArrayList<Tupla_Tabella_GO> TreeBP = new ArrayList<Tupla_Tabella_GO>();
	    Hashtable<String, ArrayList<String>> UpBP = new Hashtable<String, ArrayList<String>>();
	    Hashtable<String,ArrayList<String>> DownBP = new Hashtable<String,ArrayList<String>>();
	        
	    ArrayList<Tupla_Tabella_GO> TreeCC = new ArrayList<Tupla_Tabella_GO>();
	    Hashtable<String, ArrayList<String>> UpCC = new Hashtable<String, ArrayList<String>>();
	    Hashtable<String,ArrayList<String>> DownCC = new Hashtable<String,ArrayList<String>>();
	        
	    ArrayList<Tupla_Tabella_GO> TreeMF = new ArrayList<Tupla_Tabella_GO>();
	    Hashtable<String, ArrayList<String>> UpMF = new Hashtable<String, ArrayList<String>>();
	    Hashtable<String,ArrayList<String>> DownMF = new Hashtable<String,ArrayList<String>>();
	        
	    ArrayList<Tupla_Tabella_GO> AllGOList = new ArrayList<Tupla_Tabella_GO>();
	    
	    /* analizzo tutti i termini presenti nell'ontologia (id_OBOTerms) */
	    for(Object o:go_ontology.id_OBOTerms.keySet()){
	    	TermInfo t = go_ontology.getTerm((String) o);//prendo il termine con ID=o
	        if(t.isObsolete()==true)continue;//Filtro termini Obsoleti
	        else{
	        	if(t.getNamespace().compareTo("biological_process")==0){//filtro i termini nel dominio BP
	        		Tupla_Tabella_GO tupla = new Tupla_Tabella_GO();
	        		tupla.id=t.getId();
	        		tupla.name=t.getName();
	        		tupla.aspect=t.getNamespace();
	        		for(Object ch_ids : t.getChildren()){
	        			tupla.childs.add((String) ch_ids);
	        		}
	        		for(Object par_ids : t.getParents()){
	        			tupla.parents.add((String) par_ids);
	        		}
	        		TreeBP.add(tupla);
	        		AllGOList.add(tupla);
	        	}
	        	else if(t.getNamespace().compareTo("cellular_component")==0){//filtro i termine nel dominio CC
	        		Tupla_Tabella_GO tupla = new Tupla_Tabella_GO();
	        		tupla.id=t.getId();
	        		tupla.name=t.getName();
	        		tupla.aspect=t.getNamespace();
	        		for(Object ch_ids : t.getChildren()){
	        			tupla.childs.add((String) ch_ids);
	        		}
	        		for(Object par_ids : t.getParents()){
	        			tupla.parents.add((String) par_ids);
	        		}
	        		TreeCC.add(tupla);
	        		AllGOList.add(tupla);
	        	}
	        	else if(t.getNamespace().compareTo("molecular_function")==0){//filtro i termini nel dominio MF
	        		Tupla_Tabella_GO tupla = new Tupla_Tabella_GO();
	        		tupla.id=t.getId();
	        		tupla.name=t.getName();
	        		tupla.aspect=t.getNamespace();
	        		for(Object ch_ids : t.getChildren()){
	        			tupla.childs.add((String) ch_ids);
	        		}
	        		for(Object par_ids : t.getParents()){
	        			tupla.parents.add((String) par_ids);
	        		}
	        		TreeMF.add(tupla);
	        		AllGOList.add(tupla);
	        	}
	        }
	        	
	    }
	    /* compilo Up e Down di ogni dominio */
	    for(Tupla_Tabella_GO tr : TreeBP){
	        UpBP.put(tr.id, tr.parents);
	        DownBP.put(tr.id,tr.childs);
	    }
	    for(Tupla_Tabella_GO tr : TreeCC){
	        UpCC.put(tr.id, tr.parents);
	        DownCC.put(tr.id,tr.childs);
	    }
	    for(Tupla_Tabella_GO tr : TreeMF){
	        UpMF.put(tr.id, tr.parents);
	        DownMF.put(tr.id,tr.childs);
	    }
	    
	    
	    File directory = new File("./workspace/sim_calc/"+versione+"/");
	    if (!directory.mkdir()) {
	       System.out.println("Impossibile creare la directory "+versione+"!");
	    }
	    
	    /* salvo le strutture dati */
	    StoreFile("./workspace/sim_calc/"+versione+"/go_ontology", go_ontology);
	    //StoreFile("go_ontology23", go_ontology);
        //StoreFile("./workspace/sim_calc/"+versione+"/TreeBP", TreeBP);
	    StoreFile("./TreeBP10", TreeBP);
        StoreFile("./workspace/sim_calc/"+versione+"/UpBP", UpBP);
        StoreFile("./workspace/sim_calc/"+versione+"/DownBP", DownBP);
        
        StoreFile("./workspace/sim_calc/"+versione+"/TreeCC", TreeCC);
        StoreFile("./workspace/sim_calc/"+versione+"/UpCC", UpCC);
        StoreFile("./workspace/sim_calc/"+versione+"/DownCC", DownCC);
        
        StoreFile("./workspace/sim_calc/"+versione+"/TreeMF", TreeMF);
        StoreFile("./workspace/sim_calc/"+versione+"/UpMF", UpMF);
        StoreFile("./workspace/sim_calc/"+versione+"/DownMF", DownMF);
        
        StoreFile("./workspace/sim_calc/"+versione+"/ALLGOLIST", AllGOList);
        
        ArrayList<FileIn> lista_file = new ArrayList<FileIn>();
		ObjectInputStream ois_file;
		System.out.println("carico lista file...");
		try{
			ois_file = new ObjectInputStream(new FileInputStream("./workspace/sim_calc/lista_file"));
			lista_file = (ArrayList<FileIn>)ois_file.readObject();
		}
		catch(Exception e_f){
			System.out.println("errore caricamento lista_file");
		}
		System.out.println("fatto!");
		
		FileIn nuovo_file = new FileIn();
		
		nuovo_file.versione_OBO = versione;
		nuovo_file.organismo = "vuoto";
		ArrayList<String> lista_domini = new ArrayList<String>();
		lista_domini.add("CC");lista_domini.add("MF");lista_domini.add("BP");
		nuovo_file.domini = lista_domini;
		
		ArrayList<String> list_cod_ev = new ArrayList<String>();
		Hashtable<String, ArrayList<String>> dom_ev = new Hashtable<String, ArrayList<String>>();
		dom_ev.put("vuoto", list_cod_ev);
		nuovo_file.dom_evidence = dom_ev;
		
		lista_file.add(nuovo_file);
		//StoreFile("./workspace/sim_calc/lista_file",lista_file);
		StoreFile("lista_file",lista_file);
		
	}
	
	
	
	public static void StoreFile(String FileName, Object structure){
		System.out.println("Storing "+FileName+"...");
		ObjectOutputStream oss;
        try{
            oss = new ObjectOutputStream(new FileOutputStream(FileName));
            oss.writeObject(structure);
            oss.close();
        }
        catch(Exception e1){
                System.out.println("error during store "+FileName);
                e1.printStackTrace();
        }
        System.out.println("Done!");
	}

}
