package obo;


import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * OBOReader is a class that reads ontology files in OBO format
 * 
 * @see driver.OBOReaderDriver
 */
public class OBOReader implements Serializable
{    
    /**  name of file that contains information of terms **/
    String oboFilename;    
    /** a list of terms **/
    ArrayList OBOTerms;
    /** key: id; value: term (SlimTermInfo) **/
    public Hashtable id_OBOTerms;
    
    ArrayList<go_record> OBOTerms_senzaObsolete;

    /**
     * Constructor
     * @param oboFilename name of obo file
     */
    public OBOReader(String oboFilename)
    {
        this.OBOTerms = new ArrayList();
        this.id_OBOTerms = new Hashtable();
        this.OBOTerms_senzaObsolete = new ArrayList<go_record>();
        this.oboFilename = oboFilename;
    }
    
    /**
     * Builds the hierarchy of terms
     */
    public void buildHier()
    {
        try
        {
            FileReader fr = new FileReader(oboFilename);
            BufferedReader in = new BufferedReader(fr);
            String line;
            
            while ((line = in.readLine()) != null)
            {
                if (line.indexOf("[Term]") > -1)
                {
                	go_record rec = new go_record();
                    line = in.readLine();
                    int idPos = line.indexOf("id: ");
                    idPos += "id: ".length();
                    String id = line.substring(idPos, line.length());
                    rec.id=id;
                    
                    line = in.readLine();
                    int namePos = line.indexOf("name: ");
                    namePos += "name: ".length();
                    String name = line.substring(namePos, line.length());
                    rec.name=name;
                    
                    TermInfo term = new TermInfo(name, id);
                    
                    boolean done = false;
                    // done is true when it is done reading information about the current GO term
                    while (!done)
                    {
                        line = in.readLine();
                        if (line.length() == 0)
                            done = true;
                        
                        if (!line.contains("//"))                            
                        { 
                            // add parent for is_a relationship
                            int isaPos = line.indexOf("is_a: ");
                            if (isaPos > -1)
                            {
                                isaPos += "is_a: ".length();
                                int endPos = line.indexOf(" ", isaPos);
                                String isaID = line.substring(isaPos, endPos);
                                term.addParent(isaID);
                               
                            }
                            
                            // add parent for part_of relationship
                            int partofPos = line.indexOf("relationship: part_of ");
                            if (partofPos > -1)
                            {
                                partofPos += "relationship: part_of ".length();
                                int endPos = line.indexOf(" ", partofPos);
                                String partofID = line.substring(partofPos, endPos);
                                term.addParent(partofID);
                            }
                           
                            // add synonyms
                            int exactPos = line.indexOf("exact_synonym: ");
                            if (exactPos > -1)
                            {
                                exactPos += "exact_synonym: \"".length();
                                int endPos = line.indexOf("\"", exactPos);
                                String synonym = line.substring(exactPos, endPos);
                                term.addSynonym(synonym);
                            }
                            
                            // add synonyms
                            int relatedPos = line.indexOf("related_synonym: ");
                            if (relatedPos > -1)
                            {
                                relatedPos += "related_synonym: \"".length();
                                int endPos = line.indexOf("\"", relatedPos);
                                String synonym = line.substring(relatedPos, endPos);
                                term.addSynonym(synonym);
                            }
                            
                            // add definition
                            int defPos = line.indexOf("def: ");
                            if (defPos > -1)
                            {
                                defPos += "def: \"".length();
                                int endPos = line.indexOf("\"", defPos);
                                String def = line.substring(defPos, endPos);
                                term.setDefinition(def);
                            }
                            
                            // set true if it is obsolete
                            if (line.contains("is_obsolete: true")){
                            	term.setObsolete();
                            	rec.obsol=true;
                            }
                            else rec.obsol=false;
                            
                            if (line.startsWith("namespace: "))
                            {
                                int pos = "namespace: ".length();
                                String category = line.substring(pos, line.length());
                                term.setNamespace(category);
                                rec.tree=category;
                            }
                        }
                    }
                        
                    this.OBOTerms.add(term);
                    this.id_OBOTerms.put(id, term);
                    this.OBOTerms_senzaObsolete.add(rec);
                }
            }
            buildChildren();
        }
        catch(IOException e){ }
    }
    
    
    
    /**
     * Build the hierarchy of terms using post-order (depth-first) traversal
     *
     */
    private void buildChildren()
    {
        for (int i=0; i < this.OBOTerms.size(); i++)
        {
            TermInfo term = (TermInfo) this.OBOTerms.get(i);
            String childID = term.getId();
            ArrayList parents = term.getParents();
            for (int j=0; j < parents.size(); j++)
            {
                String parID = (String) parents.get(j);
                TermInfo parTerm = (TermInfo) this.id_OBOTerms.get(parID);
                parTerm.addChild(childID);
            }
        }
        
    }
    
    /**
     * Get a list of terms that have no children
     * @return
     */
    public ArrayList getTermsNoChild()
    {
        ArrayList terms = new ArrayList();
        for (int i=0; i < this.OBOTerms.size(); i++)
        {
            TermInfo term = (TermInfo) this.OBOTerms.get(i);
            if (term.getNumChildren() == 0)
            {
                String id = term.getId();
                terms.add(id);
            }
        }
        
        return terms;
    }
    
    /**
     * Get a list of terms that have parents
     * @return
     */
    public ArrayList getTermsWithParent()
    {
        ArrayList terms = new ArrayList();
        for (int i=0; i < this.OBOTerms.size(); i++)
        {
            TermInfo term = (TermInfo) this.OBOTerms.get(i);
            if (term.getNumParents() > 0)
            {
                String id = term.getId();
                terms.add(id);
            }
        }
        
        return terms;
    }

    /**
     * Get information about a term
     * @param index position of the GO term in the list of terms
     * @return returns infomration about the term
     */
    public TermInfo getTerm(int index)
    {
        TermInfo term = null;
        if (index > -1 && index < this.OBOTerms.size())
            term = (TermInfo) this.OBOTerms.get(index);
        return term;
    }
    
    /**
     * Get information about a term based on id
     * @param id id
     * @return returns information about the term
     */
    public TermInfo getTerm(String id)
    {
        TermInfo term = (TermInfo) this.id_OBOTerms.get(id);
        
        return term;
    }
    
    /**
     * Retrieve a list of terms that are the descendants of the term (inclusive)
     * @param term term of interest
     * @param list an empty string
     * @return a list of descendants and the term itself
     */
    public String getList(TermInfo term, String list)
    {
        ArrayList childTerms = term.getChildren();
        for (int i=0; i < term.getNumChildren(); i++)
        {
            String childTermID = (String) childTerms.get(i);
            TermInfo childTerm = (TermInfo) id_OBOTerms.get(childTermID);
            list = getList(childTerm, list); 
        }
        if(!list.contains(term.getId())){
        	list += term.getId();
        	list += "\n";
        }
        for (int i=0; i < term.getNumSynonyms(); i++)
            list += "\n" + term.getSynonym(i);
        
        return list;
    }
    
    /**
     * Retrieve a list of terms that are the ancestors of the term (inclusive)
     * @param term term of interest
     * @param list an empty string
     * @return a list of ancestors and the term itself
     */
    public String getAncestorList(TermInfo term, String list)
    {
    	//if(term.getNumParents()==0)return null;
    	ArrayList parentTerms = term.getParents();
    	
        for (int i=0; i < term.getNumParents(); i++)
        {
            String parentTermID = (String) parentTerms.get(i);
            TermInfo parentTerm = (TermInfo) id_OBOTerms.get(parentTermID);
            list = getAncestorList(parentTerm, list); 
        }
        if(!list.contains(term.getId())){
        	list += term.getId();
        	list += "\n";
        }
    	for (int i=0; i < term.getNumSynonyms(); i++)
        	list += "\n" + term.getSynonym(i);
        
        
        return list;
    	}
    
}

